# coding:utf-8
# Test for Shuffle-net, partial.
# Created   :   7,  9, 2018
# Revised   :   7,  9, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"
# os.environ['THEANO_FLAGS'] = "floatX=float32, mode=DEBUG_MODE, warn_float64='raise', exception_verbosity=high"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from dandelion.model.shufflenet import *

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def test_case_0():
    print('test_case_0: ShuffleUnit')
    model = ShuffleUnit(in_channels=16, inner_channels=4, border_mode='same', batchnorm_mode=0, activation=relu, group_num=4, fusion_mode='add', dilation=2, stride=2)
    x = tensor.ftensor4('x')
    y = model.forward(x)
    # y = tensor.nnet.conv2d()
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(4, 16, 256, 256).astype(np.float32)
    output = fn(input)
    assert output.shape == (4, 16, 128, 128), 'incorrect output shape = %s' % str(output.shape)

def test_case_1():
    print('test_case_1: ShuffleUnit_Stack')
    model = ShuffleUnit_Stack(in_channels=16, out_channels=32)
    x = tensor.ftensor4('x')
    y = model.forward(x)
    # y = tensor.nnet.conv2d()
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(4, 16, 256, 256).astype(np.float32)
    output = fn(input)
    assert output.shape == (4, 32, 128, 128), 'incorrect output shape = %s' % str(output.shape)


def test_case_2():
    print('test_case_2: model_ShuffleNet')
    model = model_ShuffleNet(in_channels=1, group_num=4, stage_channels=(24, 272, 544, 1088), stack_size=(3, 7, 3), batchnorm_mode=1, activation=relu)
    # model_weights = model.get_weights()
    # for value, w_name in model_weights:
    #     print('name = %s, shape='%w_name, value.shape)

    x = tensor.ftensor4('x')
    y = model.forward(x)
    # y = tensor.nnet.conv2d()
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(4, 1, 224, 224).astype(np.float32)
    output = fn(input)
    # print(output)
    assert output.shape == (4, 1088, 7, 7), 'incorrect output shape = %s' % str(output.shape)

def test_case_3():
    print('test_case_3: model_ShuffleSeg.ShuffleNet')
    model = model_ShuffleSeg.ShuffleNet(in_channels=1, out_channels=6)
    x = tensor.ftensor4('x')
    y = model.forward(x)
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(4, 1, 224, 224).astype(np.float32)
    output = fn(input)
    assert output[0].shape == (4, 6, 7, 7), 'incorrect output[0] shape = %s' % str(output[0].shape)
    assert output[1].shape == (4, 544, 14, 14), 'incorrect output[1] shape = %s' % str(output[1].shape)
    assert output[2].shape == (4, 272, 28, 28), 'incorrect output[2] shape = %s' % str(output[2].shape)


def test_case_4():
    print('test_case_4: model_ShuffleSeg')
    model = model_ShuffleSeg()
    # from dandelion.util import gpickle
    # gpickle.dump((model.get_weights(), None), 'shuffleseg.gpkl')
    x = tensor.ftensor4('x')
    y = model.forward(x)
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False, on_unused_input='ignore')
    print('run fn...')
    input = np.random.rand(4, 1, 256, 256).astype(np.float32)
    output = fn(input)
    assert output.shape == (4, 6, 256, 256), 'incorrect output shape = %s' % str(output.shape)

def test_case_5():
    print('test_case_5: ShuffleUnit_v2')
    args = [ [256, 128, 'same', 1, relu, 2, 2],
             [256, 256, 'same', 2, relu, 1, 1], ]
    shape_gt = [ (4, 128, 128, 128),
                 (4, 256, 256, 256)]
    for i, arg in enumerate(args):
        in_channels, out_channels, border_mode, batchnorm_mode, activation, stride, dilation = arg
        model = ShuffleUnit_v2(in_channels=in_channels, out_channels=out_channels, border_mode=border_mode, batchnorm_mode=batchnorm_mode,
                               activation=activation, dilation=dilation, stride=stride)
        x = tensor.ftensor4('x')
        y = model.forward(x)
        print('compiling fn...')
        fn = theano.function([x], y, no_default_updates=False)
        print('run fn...')
        input = np.random.rand(4, in_channels, 256, 256).astype(np.float32)
        output = fn(input)
        assert output.shape == shape_gt[i], 'incorrect output shape = %s' % str(output.shape)

def test_case_6():
    print('test_case_6: ShuffleUnit_v2_Stack')
    args = [ [16, 32, 1, relu, 3],
             [16, 16, 0, relu, 4],
             [32, 16, 2, relu, 2]]
    shape_gt = [ (4, 32, 128, 128),
                 (4, 16, 128, 128),
                 (4, 16, 128, 128)]
    for i, arg in enumerate(args):
        in_channels, out_channels, batchnorm_mode, activation, stack_size = arg
        model = ShuffleUnit_v2_Stack(in_channels=in_channels, out_channels=out_channels, batchnorm_mode=batchnorm_mode, activation=activation, stack_size=stack_size)
        x = tensor.ftensor4('x')
        y = model.forward(x)
        print('compiling fn...')
        fn = theano.function([x], y, no_default_updates=False)
        print('run fn...')
        input = np.random.rand(4, in_channels, 256, 256).astype(np.float32)
        output = fn(input)
        # print(output.shape)
        assert output.shape == shape_gt[i], 'incorrect output shape = %s' % str(output.shape)


def test_case_7():
    print('test_case_7: model_ShuffleNet_v2')
    args = [ [3, (24, 116, 232, 464, 1024), (3, 7, 3), 1, relu],
             [1, (24, 48, 96, 192, 1088), (2, 5, 2), 0, relu],
           ]
    shape_gt = [ (4, 1024, 7, 7),
                 (4, 1088, 7, 7),
                ]
    for i, arg in enumerate(args):
        in_channels, stage_channels, stack_size, batchnorm_mode, activation = arg

        model = model_ShuffleNet_v2(in_channels=in_channels, stage_channels=stage_channels, stack_size=stack_size, batchnorm_mode=batchnorm_mode, activation=activation)
        x = tensor.ftensor4('x')
        y = model.forward(x)
        # y = tensor.nnet.conv2d()
        print('compiling fn...')
        fn = theano.function([x], y, no_default_updates=False)
        print('run fn...')
        input = np.random.rand(4, in_channels, 224, 224).astype(np.float32)
        output = fn(input)
        # print(output.shape)
        assert output.shape == shape_gt[i], 'incorrect output shape = %s' % str(output.shape)

if __name__ == '__main__':

    # test_case_0()
    # test_case_1()
    # test_case_2()
    # test_case_3()
    # test_case_4()
    # test_case_5()
    # test_case_6()
    test_case_7()

    print('Test passed')






# -*- coding: utf-8 -*-
"""
Non-linear activation functions for artificial neurons.
"""

import theano.tensor as tensor

#--- element-wise activations ---#
sigmoid            = tensor.nnet.sigmoid
tanh               = tensor.tanh
relu               = tensor.nnet.relu
softplus           = tensor.nnet.softplus
ultra_fast_sigmoid = tensor.nnet.ultra_fast_sigmoid

# def _log_softmax(x):
#     xdev = x - x.max(1, keepdims=True)
#     return xdev - T.log(T.sum(T.exp(xdev), axis=1, keepdims=True))

#--- row-wise activation ---#
def log_softmax(x):
    """
    Apply log_softmax to the last dimension of input x
    :param x: tensor
    :return:
    """
    ndim = x.ndim
    if ndim <= 2:
        return tensor.nnet.logsoftmax(x)
    else:
        original_shape = x.shape
        M, N = 1, original_shape[-1]
        for i in range(x.ndim - 1):
            M = M * original_shape[i]
        x = tensor.reshape(x, (M, N))
        x = tensor.nnet.logsoftmax(x)
        x = tensor.reshape(x, original_shape)
        return x

#--- row-wise activation ---#
def softmax(x):
    """
    Apply softmax to the last dimension of input x
    :param x: tensor
    :return:
    """
    ndim = x.ndim
    if ndim <= 2:
        return tensor.nnet.softmax(x)
    else:
        original_shape = x.shape
        M, N = 1, original_shape[-1]
        for i in range(x.ndim - 1):
            M = M * original_shape[i]
        x = tensor.reshape(x, (M, N))
        x = tensor.nnet.softmax(x)
        x = tensor.reshape(x, original_shape)
        return x

# scaled tanh
class ScaledTanH(object):
    """Scaled tanh :math:`\\varphi(x) = \\tanh(\\alpha \\cdot x) \\cdot \\beta`

    This is a modified tanh function which allows to rescale both the input and
    the output of the activation.

    Scaling the input down will result in decreasing the maximum slope of the
    tanh and as a result it will be in the linear regime in a larger interval
    of the input space. Scaling the input up will increase the maximum slope
    of the tanh and thus bring it closer to a step function.

    Scaling the output changes the output interval to :math:`[-\\beta,\\beta]`.

    Parameters
    ----------
    scale_in : float32
        The scale parameter :math:`\\alpha` for the input

    scale_out : float32
        The scale parameter :math:`\\beta` for the output

    Methods
    -------
    __call__(x)
        Apply the scaled tanh function to the activation `x`.

    Examples
    --------
    In contrast to other activation functions in this module, this is
    a class that needs to be instantiated to obtain a callable:

    >>> from lasagne.layers import InputLayer, DenseLayer
    >>> l_in = InputLayer((None, 100))
    >>> from lasagne.nonlinearities import ScaledTanH
    >>> scaled_tanh = ScaledTanH(scale_in=0.5, scale_out=2.27)
    >>> l1 = DenseLayer(l_in, num_units=200, nonlinearity=scaled_tanh)

    Notes
    -----
    LeCun et al. (in [1]_, Section 4.4) suggest ``scale_in=2./3`` and
    ``scale_out=1.7159``, which has :math:`\\varphi(\\pm 1) = \\pm 1`,
    maximum second derivative at 1, and an effective gain close to 1.

    By carefully matching :math:`\\alpha` and :math:`\\beta`, the nonlinearity
    can also be tuned to preserve the mean and variance of its input:

      * ``scale_in=0.5``, ``scale_out=2.4``: If the input is a random normal
        variable, the output will have zero mean and unit variance.
      * ``scale_in=1``, ``scale_out=1.6``: Same property, but with a smaller
        linear regime in input space.
      * ``scale_in=0.5``, ``scale_out=2.27``: If the input is a uniform normal
        variable, the output will have zero mean and unit variance.
      * ``scale_in=1``, ``scale_out=1.48``: Same property, but with a smaller
        linear regime in input space.

    References
    ----------
    .. [1] LeCun, Yann A., et al. (1998):
       Efficient BackProp,
       http://link.springer.com/chapter/10.1007/3-540-49430-8_2,
       http://yann.lecun.com/exdb/publis/pdf/lecun-98b.pdf
    .. [2] Masci, Jonathan, et al. (2011):
       Stacked Convolutional Auto-Encoders for Hierarchical Feature Extraction,
       http://link.springer.com/chapter/10.1007/978-3-642-21735-7_7,
       http://people.idsia.ch/~ciresan/data/icann2011.pdf
    """

    def __init__(self, scale_in=1, scale_out=1):
        self.scale_in = scale_in
        self.scale_out = scale_out

    def __call__(self, x):
        return tensor.tanh(x * self.scale_in) * self.scale_out


ScaledTanh = ScaledTanH  # alias with alternative capitalization

# leaky rectify
class LeakyRectify(object):
    """Leaky rectifier :math:`\\varphi(x) = (x > 0)? x : \\alpha \\cdot x`

    The leaky rectifier was introduced in [1]_. Compared to the standard
    rectifier :func:`rectify`, it has a nonzero gradient for negative input,
    which often helps convergence.

    Parameters
    ----------
    leakiness : float
        Slope for negative input, usually between 0 and 1.
        A leakiness of 0 will lead to the standard rectifier,
        a leakiness of 1 will lead to a linear activation function,
        and any value in between will give a leaky rectifier.

    Methods
    -------
    __call__(x)
        Apply the leaky rectify function to the activation `x`.

    Examples
    --------
    In contrast to other activation functions in this module, this is
    a class that needs to be instantiated to obtain a callable:

    >>> from lasagne.layers import InputLayer, DenseLayer
    >>> l_in = InputLayer((None, 100))
    >>> from lasagne.nonlinearities import LeakyRectify
    >>> custom_rectify = LeakyRectify(0.1)
    >>> l1 = DenseLayer(l_in, num_units=200, nonlinearity=custom_rectify)

    Alternatively, you can use the provided instance for leakiness=0.01:

    >>> from lasagne.nonlinearities import leaky_rectify
    >>> l2 = DenseLayer(l_in, num_units=200, nonlinearity=leaky_rectify)

    Or the one for a high leakiness of 1/3:

    >>> from lasagne.nonlinearities import very_leaky_rectify
    >>> l3 = DenseLayer(l_in, num_units=200, nonlinearity=very_leaky_rectify)

    See Also
    --------
    leaky_rectify: Instance with default leakiness of 0.01, as in [1]_.
    very_leaky_rectify: Instance with high leakiness of 1/3, as in [2]_.

    References
    ----------
    .. [1] Maas et al. (2013):
       Rectifier Nonlinearities Improve Neural Network Acoustic Models,
       http://web.stanford.edu/~awni/papers/relu_hybrid_icml2013_final.pdf
    .. [2] Graham, Benjamin (2014):
       Spatially-sparse convolutional neural networks,
       http://arxiv.org/abs/1409.6070
    """
    def __init__(self, leakiness=0.01):
        self.leakiness = leakiness

    def __call__(self, x):
        return tensor.nnet.relu(x, self.leakiness)


leaky_rectify = LeakyRectify()  # shortcut with default leakiness
leaky_rectify.__doc__ = """leaky_rectify(x)

    Instance of :class:`LeakyRectify` with leakiness :math:`\\alpha=0.01`
    """


very_leaky_rectify = LeakyRectify(1./3)  # shortcut with high leakiness
very_leaky_rectify.__doc__ = """very_leaky_rectify(x)

     Instance of :class:`LeakyRectify` with leakiness :math:`\\alpha=1/3`
     """


# elu
def elu(x):
    """Exponential Linear Unit :math:`\\varphi(x) = (x > 0) ? x : e^x - 1`

    The Exponential Linear Unit (ELU) was introduced in [1]_. Compared to the
    linear rectifier :func:`rectify`, it has a mean activation closer to zero
    and nonzero gradient for negative input, which can help convergence.
    Compared to the leaky rectifier :class:`LeakyRectify`, it saturates for
    highly negative inputs.

    Parameters
    ----------
    x : float32
        The activation (the summed, weighed input of a neuron).

    Returns
    -------
    float32
        The output of the exponential linear unit for the activation.

    Notes
    -----
    In [1]_, an additional parameter :math:`\\alpha` controls the (negative)
    saturation value for negative inputs, but is set to 1 for all experiments.
    It is omitted here.

    References
    ----------
    .. [1] Djork-Arné Clevert, Thomas Unterthiner, Sepp Hochreiter (2015):
       Fast and Accurate Deep Network Learning by Exponential Linear Units
       (ELUs), http://arxiv.org/abs/1511.07289
    """
    return tensor.switch(x > 0, x, tensor.expm1(x))


# selu
class SELU(object):
    """
    Scaled Exponential Linear Unit
    :math:`\\varphi(x)=\\lambda \\left[(x>0) ? x : \\alpha(e^x-1)\\right]`

    The Scaled Exponential Linear Unit (SELU) was introduced in [1]_
    as an activation function that allows the construction of
    self-normalizing neural networks.

    Parameters
    ----------
    scale : float32
        The scale parameter :math:`\\lambda` for scaling all output.

    scale_neg  : float32
        The scale parameter :math:`\\alpha`
        for scaling output for nonpositive argument values.

    Methods
    -------
    __call__(x)
        Apply the SELU function to the activation `x`.

    Examples
    --------
    In contrast to other activation functions in this module, this is
    a class that needs to be instantiated to obtain a callable:

    >>> from lasagne.layers import InputLayer, DenseLayer
    >>> l_in = InputLayer((None, 100))
    >>> from lasagne.nonlinearities import SELU
    >>> selu = SELU(2, 3)
    >>> l1 = DenseLayer(l_in, num_units=200, nonlinearity=selu)

    See Also
    --------
    selu: Instance with :math:`\\alpha\\approx1.6733,\\lambda\\approx1.0507`
          as used in [1]_.

    References
    ----------
    .. [1] Günter Klambauer et al. (2017):
       Self-Normalizing Neural Networks,
       https://arxiv.org/abs/1706.02515
    """
    def __init__(self, scale=1, scale_neg=1):
        self.scale = scale
        self.scale_neg = scale_neg

    def __call__(self, x):
        return self.scale * tensor.switch(
                x > 0.0,
                x,
                self.scale_neg * (tensor.expm1(x)))


selu = SELU(scale=1.0507009873554804934193349852946,
            scale_neg=1.6732632423543772848170429916717)
selu.__doc__ = """selu(x)

    Instance of :class:`SELU` with :math:`\\alpha\\approx 1.6733,
    \\lambda\\approx 1.0507`

    This has a stable and attracting fixed point of :math:`\\mu=0`,
    :math:`\\sigma=1` under the assumptions of the
    original paper on self-normalizing neural networks.
    """

def linear(x):
    """Linear activation function :math:`\\varphi(x) = x`

    Parameters
    ----------
    x : float32
        The activation (the summed, weighted input of a neuron).

    Returns
    -------
    float32
        The output of the identity applied to the activation.
    """
    return x

identity = linear


# coding:utf-8
# Unit test for Dense class
# Created   :   1, 30, 2018
# Revised   :   1, 30, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from lasagne.layers import InputLayer, DenseLayer, get_output
import lasagne.nonlinearities as LACT

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

class build_model_D(Module):
    def __init__(self, in_dim=3, out_dim=3):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.dense = Dense(input_dims=self.in_dim, output_dim=self.out_dim)
        self.predict = self.forward

    def forward(self, x):
        x = self.dense.forward(x)
        x = relu(x)
        return x

def build_model_L(in_dim=3, out_dim=3):
    input_var = tensor.fmatrix('x')
    input0 = InputLayer(shape=(None, in_dim), input_var=input_var, name='input0')
    dense0 = DenseLayer(input0, num_units=out_dim, nonlinearity=LACT.rectify, name='dense0')
    return dense0

def test_case_0(in_dim=1, out_dim=1):
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name


    model_D = build_model_D(in_dim=in_dim, out_dim=out_dim)
    model_L = build_model_L(in_dim=in_dim, out_dim=out_dim)

    W = np.random.rand(in_dim, out_dim).astype(np.float32)
    b = np.random.rand(out_dim).astype(np.float32)
    model_D.dense.W.set_value(W)
    model_D.dense.b.set_value(b)
    get_layer_by_name(model_L, 'dense0').W.set_value(W)
    get_layer_by_name(model_L, 'dense0').b.set_value(b)

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True)
    fn_L = theano.function([X], y_L, no_default_updates=True)

    for i in range(20):
        x = np.random.rand(16, in_dim).astype(np.float32)
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.sum(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            raise ValueError('diff is too big')

if __name__ == '__main__':

    test_case_0(3, 2)

    print('Test passed')






# coding:utf-8
# Test VGG16 weights transferred from Lasagne
# Created   :  mm, dd, yyyy
# Revised   :  mm, dd, yyyy
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import os
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"
import theano
import lasagne
from lasagne.layers import InputLayer, get_output
from lasagne.layers import DenseLayer
from lasagne.layers import NonlinearityLayer
from lasagne.layers import DropoutLayer
from lasagne.layers import Pool2DLayer as PoolLayer
from lasagne.layers import Conv2DLayer as ConvLayer
from lasagne.nonlinearities import softmax

from dandelion.model.vgg import model_VGG16
from dandelion.util import gpickle
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)


def build_model_L():
    net = {}
    net['input'] = InputLayer((None, 3, 224, 224))
    net['conv1_1'] = ConvLayer(
        net['input'], 64, 3, pad=1, flip_filters=False)
    net['conv1_2'] = ConvLayer(
        net['conv1_1'], 64, 3, pad=1, flip_filters=False)
    net['pool1'] = PoolLayer(net['conv1_2'], 2)
    net['conv2_1'] = ConvLayer(
        net['pool1'], 128, 3, pad=1, flip_filters=False)
    net['conv2_2'] = ConvLayer(
        net['conv2_1'], 128, 3, pad=1, flip_filters=False)
    net['pool2'] = PoolLayer(net['conv2_2'], 2)
    net['conv3_1'] = ConvLayer(
        net['pool2'], 256, 3, pad=1, flip_filters=False)
    net['conv3_2'] = ConvLayer(
        net['conv3_1'], 256, 3, pad=1, flip_filters=False)
    net['conv3_3'] = ConvLayer(
        net['conv3_2'], 256, 3, pad=1, flip_filters=False)
    net['pool3'] = PoolLayer(net['conv3_3'], 2)
    net['conv4_1'] = ConvLayer(
        net['pool3'], 512, 3, pad=1, flip_filters=False)
    net['conv4_2'] = ConvLayer(
        net['conv4_1'], 512, 3, pad=1, flip_filters=False)
    net['conv4_3'] = ConvLayer(
        net['conv4_2'], 512, 3, pad=1, flip_filters=False)
    net['pool4'] = PoolLayer(net['conv4_3'], 2)
    net['conv5_1'] = ConvLayer(
        net['pool4'], 512, 3, pad=1, flip_filters=False)
    net['conv5_2'] = ConvLayer(
        net['conv5_1'], 512, 3, pad=1, flip_filters=False)
    net['conv5_3'] = ConvLayer(
        net['conv5_2'], 512, 3, pad=1, flip_filters=False)
    net['pool5'] = PoolLayer(net['conv5_3'], 2)
    net['fc6'] = DenseLayer(net['pool5'], num_units=4096)
    net['fc6_dropout'] = DropoutLayer(net['fc6'], p=0.5)
    net['fc7'] = DenseLayer(net['fc6_dropout'], num_units=4096)
    net['fc7_dropout'] = DropoutLayer(net['fc7'], p=0.5)
    net['fc8'] = DenseLayer(
        net['fc7_dropout'], num_units=1000, nonlinearity=None)
    net['prob'] = NonlinearityLayer(net['fc8'], softmax)

    return net

build_model_D = model_VGG16

# '_' prefix means no need for unit test
def _test_case_0():
    import numpy as np, pickle
    from lasagne_ext.utils import get_layer_by_name, set_weights, get_all_layers

    model_D = build_model_D()
    model_L = build_model_L()

    weight_file = r"C:\Users\dawei\Work\Code\Git\Reference Codes\Lasagne_Recipes\modelzoo\vgg16.pkl"
    with open(weight_file, mode='rb') as f:
        weights = pickle.load(f, encoding='latin1')
    lasagne.layers.set_all_param_values(model_L['prob'], weights['param values'])
    for layer_name in [
        'conv1_1',
        'conv1_2',
        'conv2_1',
        'conv2_2',
        'conv3_1',
        'conv3_2',
        'conv3_3',
        'conv4_1',
        'conv4_2',
        'conv4_3',
        'conv5_1',
        'conv5_2',
        'conv5_3',
    ]:
        print('processing layer = ', layer_name)
        W = model_L[layer_name].W.get_value()
        b = model_L[layer_name].b.get_value()
        print('W.shape=', W.shape)
        print('b.shape=', b.shape)
        model_D.__getattribute__(layer_name).W.set_value(W)
        model_D.__getattribute__(layer_name).b.set_value(b)

    for layer_name in ['fc6', 'fc7', 'fc8']:
        print('processing layer = ', layer_name)
        W = model_L[layer_name].W.get_value()
        b = model_L[layer_name].b.get_value()
        print('W.shape=', W.shape)
        print('b.shape=', b.shape)
        model_D.__getattribute__(layer_name).W.set_value(W)
        model_D.__getattribute__(layer_name).b.set_value(b)

    gpickle.dump((model_D.get_weights(), None), 'VGG16_weights.gpkl')
    print('compiling...')
    X = model_L['input'].input_var
    y_D = model_D.predict(X)
    y_L = get_output(model_L['prob'], deterministic=True)

    fn_D = theano.function([X], y_D, no_default_updates=True)
    fn_L = theano.function([X], y_L, no_default_updates=True)
    print('run test...')
    for i in range(20):
        B, C, H, W = 4, 3, 224, 224
        x = np.random.rand(B, C, H, W).astype('float32')
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.sum(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff > 1e-4:
            raise ValueError('diff is too big')









if __name__ == '__main__':

    _test_case_0()

    print('Test pass ~')
